#
# Copyright (c), 2016-2020, SISSA (International School for Advanced Studies).
# All rights reserved.
# This file is distributed under the terms of the MIT License.
# See the file 'LICENSE' in the root directory of the present
# distribution, or http://opensource.org/licenses/MIT.
#
# @author Davide Brunato <brunato@sissa.it>
#
"""
This module contains compatibility definitions.
"""
import sys
from collections import OrderedDict

ordered_dict_class = dict if sys.version_info >= (3, 6) else OrderedDict
